-module(icparse).
%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%


-export([parse/1, parse_and_scan/1, format_error/1]).

-import(lists, [reverse/1]).

-ifdef(JAM).
-compile([{parse_transform,jam_yecc_pj},pj]).
-endif.


-include("icforms.hrl").



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The parser generator will insert appropriate declarations before this line.%

parse(Tokens) ->
    case catch yeccpars1(Tokens, false, 0, [], []) of
	error ->
	    Errorline =
		if Tokens == [] -> 0; true -> element(2, hd(Tokens)) end,
	    {error,
	     {Errorline, ?MODULE, "syntax error at or after this line."}};
	Other ->
	    Other
    end.

parse_and_scan({Mod, Fun, Args}) ->
    case apply(Mod, Fun, Args) of
	{eof, _} ->
	    {ok, eof};
	{error, Descriptor, _} ->
	    {error, Descriptor};
	{ok, Tokens, _} ->
	    yeccpars1(Tokens, {Mod, Fun, Args}, 0, [], [])
    end.

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

% To be used in grammar files to throw an error message to the parser toplevel.
% Doesn't have to be exported!
return_error(Line, Message) ->
    throw({error, {Line, ?MODULE, Message}}).


% Don't change yeccpars1/6 too much, it is called recursively by yeccpars2/8!
yeccpars1([Token | Tokens], Tokenizer, State, States, Vstack) ->
    yeccpars2(State, element(1, Token), States, Vstack, Token, Tokens,
	      Tokenizer);
yeccpars1([], {M, F, A}, State, States, Vstack) ->
    case catch apply(M, F, A) of
        {eof, Endline} ->
            {error, {Endline, ?MODULE, "end_of_file"}};
        {error, Descriptor, Endline} ->
            {error, Descriptor};
        {'EXIT', Reason} ->
            {error, {0, ?MODULE, Reason}};
        {ok, Tokens, Endline} ->
	    case catch yeccpars1(Tokens, {M, F, A}, State, States, Vstack) of
		error ->
		    Errorline = element(2, hd(Tokens)),
		    {error, {Errorline, ?MODULE,
			     "syntax error at or after this line."}};
		Other ->
		    Other
	    end
    end;
yeccpars1([], false, State, States, Vstack) ->
    yeccpars2(State, '$end', States, Vstack, {'$end', 999999}, [], false).

% For internal use only.
yeccerror(Token) ->
    {error,
     {element(2, Token), ?MODULE,
      ["syntax error before: ", yecctoken2string(Token)]}}.

yecctoken2string({atom, _, A}) -> io_lib:write(A);
yecctoken2string({integer,_,N}) -> io_lib:write(N);
yecctoken2string({float,_,F}) -> io_lib:write(F);
yecctoken2string({char,_,C}) -> io_lib:write_char(C);
yecctoken2string({var,_,V}) -> io_lib:format('~s', [V]);
yecctoken2string({string,_,S}) -> io_lib:write_string(S);
yecctoken2string({reserved_symbol, _, A}) -> io_lib:format('~w', [A]);
yecctoken2string({'dot', _}) -> io_lib:format('~w', ['.']);
yecctoken2string({'$end', _}) ->
    [];
yecctoken2string({Other, _}) when atom(Other) ->
    io_lib:format('~w', [Other]);
yecctoken2string({_, _, Other}) when list(Other), number(hd(Other)) ->
    Other;
yecctoken2string({_, _, Other}) ->
    io_lib:format('~p', [Other]);
yecctoken2string(Other) ->
    io_lib:write(Other).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%-----------------------------------------------------------




yeccpars2(0, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, typedef, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, const, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, exception, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, interface, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, module, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(1, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 324, [1 | __Ss], [__T | __Stack]);
yeccpars2(1, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 323, [1 | __Ss], [__T | __Stack]);
yeccpars2(1, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(2, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 322, [2 | __Ss], [__T | __Stack]);
yeccpars2(2, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(3, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('OorM_<definition>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(4, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<type_dcl>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(5, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 321, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(6, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<interface>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(7, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 320, [7 | __Ss], [__T | __Stack]);
yeccpars2(7, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(8, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<interface>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(9, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 249, [9 | __Ss], [__T | __Stack]);
yeccpars2(9, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(10, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 248, [10 | __Ss], [__T | __Stack]);
yeccpars2(10, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(11, '$end', _, __Stack, _, _, _) ->
 {ok, hd(__Stack)};
yeccpars2(11, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(12, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<type_dcl>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(13, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 247, [13 | __Ss], [__T | __Stack]);
yeccpars2(13, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(14, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<type_dcl>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(15, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<definition>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(16, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<definition>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(17, exception, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, typedef, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, const, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, interface, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, module, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  reverse(__1),
 yeccpars2(yeccgoto('<specification>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(18, wstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, wchar, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(19, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 225, [19 | __Ss], [__T | __Stack]);
yeccpars2(19, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(20, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 219, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(21, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 211, [21 | __Ss], [__T | __Stack]);
yeccpars2(21, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(22, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 206, [22 | __Ss], [__T | __Stack]);
yeccpars2(22, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(23, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 193, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(24, wstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, sequence, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, wchar, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(25, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(26, switch, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [26 | __Ss], [__T | __Stack]);
yeccpars2(26, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(27, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(28, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, wchar, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(29, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 185, [29 | __Ss], [__T | __Stack]);
yeccpars2(29, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(30, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<switch_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(31, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<switch_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(32, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<switch_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(33, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  icgen:scoped_id_new(__1),
 yeccpars2(yeccgoto('<scoped_name>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(34, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<switch_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(35, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 109, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<switch_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(36, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<integer_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(37, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<signed_int>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(38, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<signed_int>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(39, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [39 | __Ss], [__T | __Stack]);
yeccpars2(39, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(40, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {unsigned,__1},
 yeccpars2(yeccgoto('<integer_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(41, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<unsigned_int>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(42, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<unsigned_int>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(43, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<boolean_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(44, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<char_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(45, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [45 | __Ss], [__T | __Stack]);
yeccpars2(45, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<signed_long_int>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(46, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<signed_short_int>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(47, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(48, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<char_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(49, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [49 | __Ss], [__T | __Stack]);
yeccpars2(49, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<unsigned_long_int>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(50, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<unsigned_short_int>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(51, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'long long',element(2,__2)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<unsigned_long_int>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(52, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'long long',element(2,__2)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<signed_long_int>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(53, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [53 | __Ss], [__T | __Stack]);
yeccpars2(53, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(54, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(58, __Cat, [54 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(55, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('OorM_<case>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(56, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 184, [56 | __Ss], [__T | __Stack]);
yeccpars2(56, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(57, '}', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  reverse(lists:flatten(__1)),
 yeccpars2(yeccgoto('<switch_body>', hd(__Ss)), '}', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(57, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(58, __Cat, [57 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(58, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(62, __Cat, [58 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(59, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 173, [59 | __Ss], [__T | __Stack]);
yeccpars2(59, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(60, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__1],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('Uggly_pragmas', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(61, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(117, __Cat, [61 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(62, default, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [62 | __Ss], [__T | __Stack]);
yeccpars2(62, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [62 | __Ss], [__T | __Stack]);
yeccpars2(62, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [62 | __Ss], [__T | __Stack]);
yeccpars2(62, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(63, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(116, __Cat, [63 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(64, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(65, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(66, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<case_label>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(67, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [67 | __Ss], [__T | __Stack]);
yeccpars2(67, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(68, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<unary_operator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(69, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<unary_operator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(70, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<shift_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(71, '&', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 92, [71 | __Ss], [__T | __Stack]);
yeccpars2(71, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<xor_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(72, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(73, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(74, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 113, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(75, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(76, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(77, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<primary_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(78, '%', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 100, [78 | __Ss], [__T | __Stack]);
yeccpars2(78, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [78 | __Ss], [__T | __Stack]);
yeccpars2(78, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [78 | __Ss], [__T | __Stack]);
yeccpars2(78, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<add_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(79, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 111, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_exp>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(80, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<unary_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(81, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 109, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<primary_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(82, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, '>>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<and_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(83, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(84, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<mult_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(85, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(86, '^', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<or_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(87, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<boolean_literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(88, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<boolean_literal>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(89, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<unary_operator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(90, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(91, '&', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 92, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'xor',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<xor_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(92, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(93, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '>>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'and',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<and_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(94, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(95, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(96, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {rshift,__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<shift_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(97, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(98, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(99, '%', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 100, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'-',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<add_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(100, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(101, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(102, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(103, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'/',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<mult_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(104, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'*',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<mult_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(105, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'%',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<mult_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(106, '%', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 100, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'+',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<add_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(107, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {lshift,__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<shift_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(108, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<unary_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(109, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 110, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(110, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  icgen:scoped_id_add(__1,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<scoped_name>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(111, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(112, '^', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [112 | __Ss], [__T | __Stack]);
yeccpars2(112, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'or',__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<or_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(113, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<case_label>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(114, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(115, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<primary_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(116, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1 ++ ([__2] ++ __3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('OorM_<case_label>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(117, wstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, sequence, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, default, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, wchar, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(118, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(119, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<simple_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(120, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(121, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(172, __Cat, [121 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(122, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(123, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(124, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(169, __Cat, [124 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(125, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<constr_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(126, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(127, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(128, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(129, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 109, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<simple_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(130, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<template_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(131, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(132, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<template_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(133, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<constr_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(134, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<simple_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(135, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 161, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(136, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<constr_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(137, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<base_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(138, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<any_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(139, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<floating_pt_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(140, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<floating_pt_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(141, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<octet_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(142, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 152, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(143, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 149, [143 | __Ss], [__T | __Stack]);
yeccpars2(143, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #string{},
 yeccpars2(yeccgoto('<string_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(144, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #wstring{},
 yeccpars2(yeccgoto('<string_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(145, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(146, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<positive_int_const>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(147, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(148, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #wstring{length = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('<string_type>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(149, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(150, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 151, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(151, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #string{length = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('<string_type>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(152, wstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, sequence, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, wchar, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(153, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 155, [153 | __Ss], [__T | __Stack]);
yeccpars2(153, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 154, [153 | __Ss], [__T | __Stack]);
yeccpars2(153, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(154, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(155, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #sequence{type = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('<sequence_type>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(156, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 157, [156 | __Ss], [__T | __Stack]);
yeccpars2(156, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(157, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #sequence{type = __3, length = __5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('<sequence_type>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(158, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<complex_declarator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(159, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<declarator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(160, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #case_dcl{type = __1, id = __2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<element_spec>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(161, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [161 | __Ss], [__T | __Stack]);
yeccpars2(161, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<simple_declarator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(162, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<declarator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(163, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto('OorM_<fixed_array_size>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(164, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #array{id = __1, size = reverse(__2)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<array_declarator>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(165, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(166, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 167, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(167, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<fixed_array_size>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(168, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__1],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('OorM_<fixed_array_size>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(169, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 170, [169 | __Ss], [__T | __Stack]);
yeccpars2(169, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [169 | __Ss], [__T | __Stack]);
yeccpars2(169, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(170, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(171, __Cat, [170 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(171, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1 ++ (__3 ++ (__5 ++ (__7 ++ [__4#case_dcl{label = reverse(__2)}]))),
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto('<case>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(172, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 ++ ([__3|__1] ++ __4),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('OorM_<case_label>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(173, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 174, [173 | __Ss], [__T | __Stack]);
yeccpars2(173, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(174, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 175, [174 | __Ss], [__T | __Stack]);
yeccpars2(174, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(175, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 176, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 177, [175 | __Ss], [__T | __Stack]);
yeccpars2(175, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(176, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 179, [176 | __Ss], [__T | __Stack]);
yeccpars2(176, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 180, [176 | __Ss], [__T | __Stack]);
yeccpars2(176, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(177, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 178, [177 | __Ss], [__T | __Stack]);
yeccpars2(177, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(178, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #pragma{type = __4, to = followed, apply = __5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('OE_pragma', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(179, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 182, [179 | __Ss], [__T | __Stack]);
yeccpars2(179, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(180, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 181, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(181, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #pragma{type = __4, to = __5, apply = __6},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto('OE_pragma', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(182, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #pragma{type = __4, to = __5, apply = icgen:float_to_version(__6)},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto('OE_pragma', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(183, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__1],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('OorM_<case>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(184, __Cat, __Ss,  [__9,__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #union{id = __2, type = __5, body = __8},
 __Nss = lists:nthtail(8, __Ss),
 yeccpars2(yeccgoto('<union_type>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(185, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  icgen:scoped_id_new_global(__2),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<scoped_name>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(186, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<type_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(187, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 161, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(188, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(190, __Cat, [188 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(189, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #typedef{type = __1, id = __2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<type_declarator>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(190, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 191, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|reverse(__2)],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<declarators>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(191, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 161, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(192, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__1],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ZorM_<declarator>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(193, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 194, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(194, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(198, __Cat, [194 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(195, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('OorM_<member>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(196, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 205, [196 | __Ss], [__T | __Stack]);
yeccpars2(196, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(197, '}', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  reverse(__1),
 yeccpars2(yeccgoto('<member_list>', hd(__Ss)), '}', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(197, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(198, __Cat, [197 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(198, wstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, sequence, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, wchar, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(199, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 161, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(200, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(201, __Cat, [200 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(201, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 202, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(202, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(203, __Cat, [202 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(203, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1 ++ (__4 ++ (__6 ++ [#member{type = __2, id = __3}])),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('<member>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(204, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 ++ __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('OorM_<member>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(205, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #struct{id = __2, body = __4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<struct_type>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(206, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 207, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(207, exception, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, typedef, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, const, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, interface, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, module, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(208, exception, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, typedef, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, const, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, interface, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 210, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, module, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(209, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__1],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('OorM_<definition>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(210, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #module{id = __2, body = reverse(__4)},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<module>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(211, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 212, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, ';', __Ss, [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #forward{id = __2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<forward_dcl>', hd(__Nss)), ';', __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(211, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(214, __Cat, [211 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(212, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(213, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Opt_<inheritance_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(214, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<interface_header>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(215, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 109, [215 | __Ss], [__T | __Stack]);
yeccpars2(215, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(216, __Cat, [215 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(216, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 217, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|reverse(__3)],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<inheritance_spec>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(217, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(218, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 109, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__1],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ZorM_<scoped_name>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(219, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 220, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(220, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(221, __Cat, [220 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(221, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ZorM_<member>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(222, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 224, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(198, __Cat, [222 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(223, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 ++ __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ZorM_<member>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(224, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #except{id = __2, body = reverse(__4)},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<except_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(225, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 226, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(226, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(227, __Cat, [226 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(227, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 229, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(228, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(230, __Cat, [228 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(229, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #enumerator{id = __1},
 yeccpars2(yeccgoto('<enumerator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(230, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(231, __Cat, [230 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(231, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(232, __Cat, [231 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(232, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 233, [232 | __Ss], [__T | __Stack]);
yeccpars2(232, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 234, [232 | __Ss], [__T | __Stack]);
yeccpars2(232, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [232 | __Ss], [__T | __Stack]);
yeccpars2(232, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(233, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(235, __Cat, [233 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(234, __Cat, __Ss,  [__9,__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #enum{id = __2, body = __4 ++ (__6 ++ (__8 ++ [__5|reverse(__7)]))},
 __Nss = lists:nthtail(8, __Ss),
 yeccpars2(yeccgoto('<enum_type>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(235, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 229, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(236, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 ++ (__4 ++ [__5|__1]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('ZorM_<enumerator>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(237, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(238, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(239, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 244, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(240, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(241, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(242, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 109, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(243, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<const_type>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(244, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 245, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(245, 'FALSE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, 'TRUE', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, '<floating_pt_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, '<character_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, '~', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(246, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #const{type = __2, id = __3, val = __5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<const_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(247, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<definition>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(248, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<definition>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(249, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(251, __Cat, [249 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(250, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 319, [250 | __Ss], [__T | __Stack]);
yeccpars2(250, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(251, readonly, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 264, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, oneway, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 263, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto('<interface_body>', hd(__Ss)), '}', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(251, attribute, __Ss, __Stack, __T, __Ts, __Tzr) ->
 __Val =  nil,
 yeccpars2(262, attribute, [251 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(251, exception, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, enum, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, union, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, struct, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, typedef, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, const, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  nil,
 yeccpars2(261, __Cat, [251 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(252, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 318, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(253, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 317, [253 | __Ss], [__T | __Stack]);
yeccpars2(253, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(254, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 316, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(255, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  if list(__2) -> __2 ++ __1; true -> [__2|__1] end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ZorM_<export>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(256, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Opt_<op_attribute>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(257, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 315, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(258, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 314, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(259, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<export>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(260, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<export>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(261, void, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 277, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, wstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, wchar, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(262, attribute, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 265, [262 | __Ss], [__T | __Stack]);
yeccpars2(262, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(263, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<op_attribute>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(264, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Opt_readonly', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(265, wstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, wchar, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(266, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<param_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(267, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 270, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(268, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 109, [268 | __Ss], [__T | __Stack]);
yeccpars2(268, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<param_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(269, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<param_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(270, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<simple_declarator>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(271, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(272, __Cat, [271 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(272, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 273, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #attr{readonly = __1, type = __3, id = [__4|reverse(__5)]},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<attr_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(273, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 270, [273 | __Ss], [__T | __Stack]);
yeccpars2(273, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(274, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__1],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ZorM_<simple_declarator>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(275, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 278, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(276, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<op_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(277, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<op_type_spec>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(278, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 279, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(279, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(297, __Cat, [279 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(280, raises, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 283, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(282, __Cat, [280 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(281, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Opt_<raises_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(282, context, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 290, [282 | __Ss], [__T | __Stack]);
yeccpars2(282, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(289, __Cat, [282 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(283, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 284, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(284, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(285, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 109, [285 | __Ss], [__T | __Stack]);
yeccpars2(285, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(286, __Cat, [285 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(286, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 287, [286 | __Ss], [__T | __Stack]);
yeccpars2(286, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 217, [286 | __Ss], [__T | __Stack]);
yeccpars2(286, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(287, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|reverse(__4)],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<raises_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(288, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('Opt_<context_expr>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(289, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #op{oneway = __1, type = __2, id = __3, params = __4, raises = __5, ctx = __6},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('<op_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(290, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 291, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(291, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 292, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(292, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(293, __Cat, [292 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(293, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 295, [293 | __Ss], [__T | __Stack]);
yeccpars2(293, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 294, [293 | __Ss], [__T | __Stack]);
yeccpars2(293, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(294, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|reverse(__4)],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<context_expr>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(295, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 296, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(296, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__1],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('ZorM_<string_literal>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(297, inout, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 302, [297 | __Ss], [__T | __Stack]);
yeccpars2(297, out, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 303, [297 | __Ss], [__T | __Stack]);
yeccpars2(297, in, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 301, [297 | __Ss], [__T | __Stack]);
yeccpars2(297, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 298, [297 | __Ss], [__T | __Stack]);
yeccpars2(297, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [297 | __Ss], [__T | __Stack]);
yeccpars2(297, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(298, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<parameter_dcls>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(299, wstring, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, any, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, octet, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, boolean, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, wchar, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, unsigned, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, short, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, long, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, double, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, 'Object', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, '::', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(300, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(304, __Cat, [300 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(301, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<param_attribute>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(302, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<param_attribute>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(303, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('<param_attribute>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(304, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto('ZorM_<param_dcl>', hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(305, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 306, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(307, __Cat, [305 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(306, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 ++ [__3|reverse(__4)],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('<parameter_dcls>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(307, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 308, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(308, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(309, __Cat, [308 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(309, inout, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 302, [309 | __Ss], [__T | __Stack]);
yeccpars2(309, out, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 303, [309 | __Ss], [__T | __Stack]);
yeccpars2(309, in, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 301, [309 | __Ss], [__T | __Stack]);
yeccpars2(309, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [309 | __Ss], [__T | __Stack]);
yeccpars2(309, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(310, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(311, __Cat, [310 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(311, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2 ++ (__4 ++ (__6 ++ [__5|__1])),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto('ZorM_<param_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(312, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 270, [312 | __Ss], [__T | __Stack]);
yeccpars2(312, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(313, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #param{inout = __1, type = __2, id = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto('<param_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(314, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<export>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(315, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<export>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(316, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<export>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(317, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<export>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(318, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<export>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(319, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #interface{id = element(1,__1), inherit = element(2,__1), body = lists:reverse(__3)},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto('<interface_dcl>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(320, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<definition>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(321, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<definition>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(322, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('<definition>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(323, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  '$undefined',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('OE_preproc', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(324, '<string_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 325, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, '<identifier>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 174, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(325, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 326, [325 | __Ss], [__T | __Stack]);
yeccpars2(325, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(327, __Cat, [325 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(326, '<integer_literal>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 326, [326 | __Ss], [__T | __Stack]);
yeccpars2(326, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(329, __Cat, [326 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(327, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 328, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(328, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  case __4 of [] -> case __2 of {_,_,"1"} -> #preproc{cat = line_nr, id = __3, aux = __4}; _ -> [] end; _ -> #preproc{cat = line_nr, id = __3, aux = __4} end,
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto('OE_preproc', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(329, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto('ZorM_<integer_literal>', hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(__Other, _, _, _, _, _, _) ->
 exit({parser, __Other, missing_state_in_action_table}).

yeccgoto('<add_expr>', 64) ->
 70;
yeccgoto('<add_expr>', 67) ->
 70;
yeccgoto('<add_expr>', 90) ->
 70;
yeccgoto('<add_expr>', 92) ->
 70;
yeccgoto('<add_expr>', 94) ->
 107;
yeccgoto('<add_expr>', 95) ->
 96;
yeccgoto('<add_expr>', 111) ->
 70;
yeccgoto('<add_expr>', 145) ->
 70;
yeccgoto('<add_expr>', 149) ->
 70;
yeccgoto('<add_expr>', 154) ->
 70;
yeccgoto('<add_expr>', 165) ->
 70;
yeccgoto('<add_expr>', 245) ->
 70;
yeccgoto('<and_expr>', 64) ->
 71;
yeccgoto('<and_expr>', 67) ->
 71;
yeccgoto('<and_expr>', 90) ->
 91;
yeccgoto('<and_expr>', 111) ->
 71;
yeccgoto('<and_expr>', 145) ->
 71;
yeccgoto('<and_expr>', 149) ->
 71;
yeccgoto('<and_expr>', 154) ->
 71;
yeccgoto('<and_expr>', 165) ->
 71;
yeccgoto('<and_expr>', 245) ->
 71;
yeccgoto('<any_type>', 24) ->
 118;
yeccgoto('<any_type>', 117) ->
 118;
yeccgoto('<any_type>', 152) ->
 118;
yeccgoto('<any_type>', 198) ->
 118;
yeccgoto('<any_type>', 261) ->
 118;
yeccgoto('<any_type>', 265) ->
 118;
yeccgoto('<any_type>', 299) ->
 118;
yeccgoto('<array_declarator>', 135) ->
 158;
yeccgoto('<array_declarator>', 187) ->
 158;
yeccgoto('<array_declarator>', 191) ->
 158;
yeccgoto('<array_declarator>', 199) ->
 158;
yeccgoto('<attr_dcl>', 251) ->
 252;
yeccgoto('<base_type_spec>', 24) ->
 119;
yeccgoto('<base_type_spec>', 117) ->
 119;
yeccgoto('<base_type_spec>', 152) ->
 119;
yeccgoto('<base_type_spec>', 198) ->
 119;
yeccgoto('<base_type_spec>', 261) ->
 266;
yeccgoto('<base_type_spec>', 265) ->
 266;
yeccgoto('<base_type_spec>', 299) ->
 266;
yeccgoto('<boolean_literal>', 64) ->
 72;
yeccgoto('<boolean_literal>', 67) ->
 72;
yeccgoto('<boolean_literal>', 85) ->
 72;
yeccgoto('<boolean_literal>', 90) ->
 72;
yeccgoto('<boolean_literal>', 92) ->
 72;
yeccgoto('<boolean_literal>', 94) ->
 72;
yeccgoto('<boolean_literal>', 95) ->
 72;
yeccgoto('<boolean_literal>', 97) ->
 72;
yeccgoto('<boolean_literal>', 98) ->
 72;
yeccgoto('<boolean_literal>', 100) ->
 72;
yeccgoto('<boolean_literal>', 101) ->
 72;
yeccgoto('<boolean_literal>', 102) ->
 72;
yeccgoto('<boolean_literal>', 111) ->
 72;
yeccgoto('<boolean_literal>', 145) ->
 72;
yeccgoto('<boolean_literal>', 149) ->
 72;
yeccgoto('<boolean_literal>', 154) ->
 72;
yeccgoto('<boolean_literal>', 165) ->
 72;
yeccgoto('<boolean_literal>', 245) ->
 72;
yeccgoto('<boolean_type>', 18) ->
 237;
yeccgoto('<boolean_type>', 24) ->
 120;
yeccgoto('<boolean_type>', 28) ->
 30;
yeccgoto('<boolean_type>', 117) ->
 120;
yeccgoto('<boolean_type>', 152) ->
 120;
yeccgoto('<boolean_type>', 198) ->
 120;
yeccgoto('<boolean_type>', 261) ->
 120;
yeccgoto('<boolean_type>', 265) ->
 120;
yeccgoto('<boolean_type>', 299) ->
 120;
yeccgoto('<case>', 54) ->
 55;
yeccgoto('<case>', 57) ->
 183;
yeccgoto('<case_label>', 62) ->
 63;
yeccgoto('<case_label>', 117) ->
 121;
yeccgoto('<char_type>', 18) ->
 238;
yeccgoto('<char_type>', 24) ->
 122;
yeccgoto('<char_type>', 28) ->
 31;
yeccgoto('<char_type>', 117) ->
 122;
yeccgoto('<char_type>', 152) ->
 122;
yeccgoto('<char_type>', 198) ->
 122;
yeccgoto('<char_type>', 261) ->
 122;
yeccgoto('<char_type>', 265) ->
 122;
yeccgoto('<char_type>', 299) ->
 122;
yeccgoto('<complex_declarator>', 135) ->
 159;
yeccgoto('<complex_declarator>', 187) ->
 159;
yeccgoto('<complex_declarator>', 191) ->
 159;
yeccgoto('<complex_declarator>', 199) ->
 159;
yeccgoto('<const_dcl>', 0) ->
 2;
yeccgoto('<const_dcl>', 17) ->
 2;
yeccgoto('<const_dcl>', 207) ->
 2;
yeccgoto('<const_dcl>', 208) ->
 2;
yeccgoto('<const_dcl>', 251) ->
 253;
yeccgoto('<const_exp>', 64) ->
 74;
yeccgoto('<const_exp>', 67) ->
 114;
yeccgoto('<const_exp>', 145) ->
 146;
yeccgoto('<const_exp>', 149) ->
 146;
yeccgoto('<const_exp>', 154) ->
 146;
yeccgoto('<const_exp>', 165) ->
 146;
yeccgoto('<const_exp>', 245) ->
 246;
yeccgoto('<const_type>', 18) ->
 239;
yeccgoto('<constr_type_spec>', 24) ->
 123;
yeccgoto('<constr_type_spec>', 117) ->
 123;
yeccgoto('<constr_type_spec>', 198) ->
 123;
yeccgoto('<context_expr>', 282) ->
 288;
yeccgoto('<declarator>', 135) ->
 160;
yeccgoto('<declarator>', 187) ->
 188;
yeccgoto('<declarator>', 191) ->
 192;
yeccgoto('<declarator>', 199) ->
 188;
yeccgoto('<declarators>', 187) ->
 189;
yeccgoto('<declarators>', 199) ->
 200;
yeccgoto('<definition>', 0) ->
 3;
yeccgoto('<definition>', 17) ->
 209;
yeccgoto('<definition>', 207) ->
 3;
yeccgoto('<definition>', 208) ->
 209;
yeccgoto('<element_spec>', 117) ->
 124;
yeccgoto('<enum_type>', 0) ->
 4;
yeccgoto('<enum_type>', 17) ->
 4;
yeccgoto('<enum_type>', 24) ->
 125;
yeccgoto('<enum_type>', 28) ->
 32;
yeccgoto('<enum_type>', 117) ->
 125;
yeccgoto('<enum_type>', 198) ->
 125;
yeccgoto('<enum_type>', 207) ->
 4;
yeccgoto('<enum_type>', 208) ->
 4;
yeccgoto('<enum_type>', 251) ->
 4;
yeccgoto('<enumerator>', 227) ->
 228;
yeccgoto('<enumerator>', 235) ->
 236;
yeccgoto('<except_dcl>', 0) ->
 5;
yeccgoto('<except_dcl>', 17) ->
 5;
yeccgoto('<except_dcl>', 207) ->
 5;
yeccgoto('<except_dcl>', 208) ->
 5;
yeccgoto('<except_dcl>', 251) ->
 254;
yeccgoto('<export>', 251) ->
 255;
yeccgoto('<fixed_array_size>', 161) ->
 163;
yeccgoto('<fixed_array_size>', 164) ->
 168;
yeccgoto('<floating_pt_type>', 18) ->
 240;
yeccgoto('<floating_pt_type>', 24) ->
 126;
yeccgoto('<floating_pt_type>', 117) ->
 126;
yeccgoto('<floating_pt_type>', 152) ->
 126;
yeccgoto('<floating_pt_type>', 198) ->
 126;
yeccgoto('<floating_pt_type>', 261) ->
 126;
yeccgoto('<floating_pt_type>', 265) ->
 126;
yeccgoto('<floating_pt_type>', 299) ->
 126;
yeccgoto('<forward_dcl>', 0) ->
 6;
yeccgoto('<forward_dcl>', 17) ->
 6;
yeccgoto('<forward_dcl>', 207) ->
 6;
yeccgoto('<forward_dcl>', 208) ->
 6;
yeccgoto('<inheritance_spec>', 211) ->
 213;
yeccgoto('<integer_type>', 18) ->
 241;
yeccgoto('<integer_type>', 24) ->
 127;
yeccgoto('<integer_type>', 28) ->
 34;
yeccgoto('<integer_type>', 117) ->
 127;
yeccgoto('<integer_type>', 152) ->
 127;
yeccgoto('<integer_type>', 198) ->
 127;
yeccgoto('<integer_type>', 261) ->
 127;
yeccgoto('<integer_type>', 265) ->
 127;
yeccgoto('<integer_type>', 299) ->
 127;
yeccgoto('<interface>', 0) ->
 7;
yeccgoto('<interface>', 17) ->
 7;
yeccgoto('<interface>', 207) ->
 7;
yeccgoto('<interface>', 208) ->
 7;
yeccgoto('<interface_body>', 249) ->
 250;
yeccgoto('<interface_dcl>', 0) ->
 8;
yeccgoto('<interface_dcl>', 17) ->
 8;
yeccgoto('<interface_dcl>', 207) ->
 8;
yeccgoto('<interface_dcl>', 208) ->
 8;
yeccgoto('<interface_header>', 0) ->
 9;
yeccgoto('<interface_header>', 17) ->
 9;
yeccgoto('<interface_header>', 207) ->
 9;
yeccgoto('<interface_header>', 208) ->
 9;
yeccgoto('<literal>', 64) ->
 77;
yeccgoto('<literal>', 67) ->
 77;
yeccgoto('<literal>', 85) ->
 77;
yeccgoto('<literal>', 90) ->
 77;
yeccgoto('<literal>', 92) ->
 77;
yeccgoto('<literal>', 94) ->
 77;
yeccgoto('<literal>', 95) ->
 77;
yeccgoto('<literal>', 97) ->
 77;
yeccgoto('<literal>', 98) ->
 77;
yeccgoto('<literal>', 100) ->
 77;
yeccgoto('<literal>', 101) ->
 77;
yeccgoto('<literal>', 102) ->
 77;
yeccgoto('<literal>', 111) ->
 77;
yeccgoto('<literal>', 145) ->
 77;
yeccgoto('<literal>', 149) ->
 77;
yeccgoto('<literal>', 154) ->
 77;
yeccgoto('<literal>', 165) ->
 77;
yeccgoto('<literal>', 245) ->
 77;
yeccgoto('<member>', 194) ->
 195;
yeccgoto('<member>', 197) ->
 204;
yeccgoto('<member>', 222) ->
 223;
yeccgoto('<member_list>', 194) ->
 196;
yeccgoto('<module>', 0) ->
 10;
yeccgoto('<module>', 17) ->
 10;
yeccgoto('<module>', 207) ->
 10;
yeccgoto('<module>', 208) ->
 10;
yeccgoto('<mult_expr>', 64) ->
 78;
yeccgoto('<mult_expr>', 67) ->
 78;
yeccgoto('<mult_expr>', 90) ->
 78;
yeccgoto('<mult_expr>', 92) ->
 78;
yeccgoto('<mult_expr>', 94) ->
 78;
yeccgoto('<mult_expr>', 95) ->
 78;
yeccgoto('<mult_expr>', 97) ->
 106;
yeccgoto('<mult_expr>', 98) ->
 99;
yeccgoto('<mult_expr>', 111) ->
 78;
yeccgoto('<mult_expr>', 145) ->
 78;
yeccgoto('<mult_expr>', 149) ->
 78;
yeccgoto('<mult_expr>', 154) ->
 78;
yeccgoto('<mult_expr>', 165) ->
 78;
yeccgoto('<mult_expr>', 245) ->
 78;
yeccgoto('<octet_type>', 24) ->
 128;
yeccgoto('<octet_type>', 117) ->
 128;
yeccgoto('<octet_type>', 152) ->
 128;
yeccgoto('<octet_type>', 198) ->
 128;
yeccgoto('<octet_type>', 261) ->
 128;
yeccgoto('<octet_type>', 265) ->
 128;
yeccgoto('<octet_type>', 299) ->
 128;
yeccgoto('<op_attribute>', 251) ->
 256;
yeccgoto('<op_dcl>', 251) ->
 257;
yeccgoto('<op_type_spec>', 261) ->
 275;
yeccgoto('<or_expr>', 64) ->
 79;
yeccgoto('<or_expr>', 67) ->
 79;
yeccgoto('<or_expr>', 145) ->
 79;
yeccgoto('<or_expr>', 149) ->
 79;
yeccgoto('<or_expr>', 154) ->
 79;
yeccgoto('<or_expr>', 165) ->
 79;
yeccgoto('<or_expr>', 245) ->
 79;
yeccgoto('<param_attribute>', 297) ->
 299;
yeccgoto('<param_attribute>', 309) ->
 299;
yeccgoto('<param_dcl>', 297) ->
 300;
yeccgoto('<param_dcl>', 309) ->
 310;
yeccgoto('<param_type_spec>', 261) ->
 276;
yeccgoto('<param_type_spec>', 265) ->
 267;
yeccgoto('<param_type_spec>', 299) ->
 312;
yeccgoto('<parameter_dcls>', 278) ->
 280;
yeccgoto('<positive_int_const>', 145) ->
 147;
yeccgoto('<positive_int_const>', 149) ->
 150;
yeccgoto('<positive_int_const>', 154) ->
 156;
yeccgoto('<positive_int_const>', 165) ->
 166;
yeccgoto('<primary_expr>', 64) ->
 80;
yeccgoto('<primary_expr>', 67) ->
 80;
yeccgoto('<primary_expr>', 85) ->
 108;
yeccgoto('<primary_expr>', 90) ->
 80;
yeccgoto('<primary_expr>', 92) ->
 80;
yeccgoto('<primary_expr>', 94) ->
 80;
yeccgoto('<primary_expr>', 95) ->
 80;
yeccgoto('<primary_expr>', 97) ->
 80;
yeccgoto('<primary_expr>', 98) ->
 80;
yeccgoto('<primary_expr>', 100) ->
 80;
yeccgoto('<primary_expr>', 101) ->
 80;
yeccgoto('<primary_expr>', 102) ->
 80;
yeccgoto('<primary_expr>', 111) ->
 80;
yeccgoto('<primary_expr>', 145) ->
 80;
yeccgoto('<primary_expr>', 149) ->
 80;
yeccgoto('<primary_expr>', 154) ->
 80;
yeccgoto('<primary_expr>', 165) ->
 80;
yeccgoto('<primary_expr>', 245) ->
 80;
yeccgoto('<raises_expr>', 280) ->
 281;
yeccgoto('<scoped_name>', 18) ->
 242;
yeccgoto('<scoped_name>', 24) ->
 129;
yeccgoto('<scoped_name>', 28) ->
 35;
yeccgoto('<scoped_name>', 64) ->
 81;
yeccgoto('<scoped_name>', 67) ->
 81;
yeccgoto('<scoped_name>', 85) ->
 81;
yeccgoto('<scoped_name>', 90) ->
 81;
yeccgoto('<scoped_name>', 92) ->
 81;
yeccgoto('<scoped_name>', 94) ->
 81;
yeccgoto('<scoped_name>', 95) ->
 81;
yeccgoto('<scoped_name>', 97) ->
 81;
yeccgoto('<scoped_name>', 98) ->
 81;
yeccgoto('<scoped_name>', 100) ->
 81;
yeccgoto('<scoped_name>', 101) ->
 81;
yeccgoto('<scoped_name>', 102) ->
 81;
yeccgoto('<scoped_name>', 111) ->
 81;
yeccgoto('<scoped_name>', 117) ->
 129;
yeccgoto('<scoped_name>', 145) ->
 81;
yeccgoto('<scoped_name>', 149) ->
 81;
yeccgoto('<scoped_name>', 152) ->
 129;
yeccgoto('<scoped_name>', 154) ->
 81;
yeccgoto('<scoped_name>', 165) ->
 81;
yeccgoto('<scoped_name>', 198) ->
 129;
yeccgoto('<scoped_name>', 212) ->
 215;
yeccgoto('<scoped_name>', 217) ->
 218;
yeccgoto('<scoped_name>', 245) ->
 81;
yeccgoto('<scoped_name>', 261) ->
 268;
yeccgoto('<scoped_name>', 265) ->
 268;
yeccgoto('<scoped_name>', 284) ->
 285;
yeccgoto('<scoped_name>', 299) ->
 268;
yeccgoto('<sequence_type>', 24) ->
 130;
yeccgoto('<sequence_type>', 117) ->
 130;
yeccgoto('<sequence_type>', 152) ->
 130;
yeccgoto('<sequence_type>', 198) ->
 130;
yeccgoto('<shift_expr>', 64) ->
 82;
yeccgoto('<shift_expr>', 67) ->
 82;
yeccgoto('<shift_expr>', 90) ->
 82;
yeccgoto('<shift_expr>', 92) ->
 93;
yeccgoto('<shift_expr>', 111) ->
 82;
yeccgoto('<shift_expr>', 145) ->
 82;
yeccgoto('<shift_expr>', 149) ->
 82;
yeccgoto('<shift_expr>', 154) ->
 82;
yeccgoto('<shift_expr>', 165) ->
 82;
yeccgoto('<shift_expr>', 245) ->
 82;
yeccgoto('<signed_int>', 18) ->
 36;
yeccgoto('<signed_int>', 24) ->
 36;
yeccgoto('<signed_int>', 28) ->
 36;
yeccgoto('<signed_int>', 117) ->
 36;
yeccgoto('<signed_int>', 152) ->
 36;
yeccgoto('<signed_int>', 198) ->
 36;
yeccgoto('<signed_int>', 261) ->
 36;
yeccgoto('<signed_int>', 265) ->
 36;
yeccgoto('<signed_int>', 299) ->
 36;
yeccgoto('<signed_long_int>', 18) ->
 37;
yeccgoto('<signed_long_int>', 24) ->
 37;
yeccgoto('<signed_long_int>', 28) ->
 37;
yeccgoto('<signed_long_int>', 117) ->
 37;
yeccgoto('<signed_long_int>', 152) ->
 37;
yeccgoto('<signed_long_int>', 198) ->
 37;
yeccgoto('<signed_long_int>', 261) ->
 37;
yeccgoto('<signed_long_int>', 265) ->
 37;
yeccgoto('<signed_long_int>', 299) ->
 37;
yeccgoto('<signed_short_int>', 18) ->
 38;
yeccgoto('<signed_short_int>', 24) ->
 38;
yeccgoto('<signed_short_int>', 28) ->
 38;
yeccgoto('<signed_short_int>', 117) ->
 38;
yeccgoto('<signed_short_int>', 152) ->
 38;
yeccgoto('<signed_short_int>', 198) ->
 38;
yeccgoto('<signed_short_int>', 261) ->
 38;
yeccgoto('<signed_short_int>', 265) ->
 38;
yeccgoto('<signed_short_int>', 299) ->
 38;
yeccgoto('<simple_declarator>', 135) ->
 162;
yeccgoto('<simple_declarator>', 187) ->
 162;
yeccgoto('<simple_declarator>', 191) ->
 162;
yeccgoto('<simple_declarator>', 199) ->
 162;
yeccgoto('<simple_declarator>', 267) ->
 271;
yeccgoto('<simple_declarator>', 273) ->
 274;
yeccgoto('<simple_declarator>', 312) ->
 313;
yeccgoto('<simple_type_spec>', 24) ->
 131;
yeccgoto('<simple_type_spec>', 117) ->
 131;
yeccgoto('<simple_type_spec>', 152) ->
 153;
yeccgoto('<simple_type_spec>', 198) ->
 131;
yeccgoto('<specification>', 0) ->
 11;
yeccgoto('<string_type>', 18) ->
 243;
yeccgoto('<string_type>', 24) ->
 132;
yeccgoto('<string_type>', 117) ->
 132;
yeccgoto('<string_type>', 152) ->
 132;
yeccgoto('<string_type>', 198) ->
 132;
yeccgoto('<string_type>', 261) ->
 269;
yeccgoto('<string_type>', 265) ->
 269;
yeccgoto('<string_type>', 299) ->
 269;
yeccgoto('<struct_type>', 0) ->
 12;
yeccgoto('<struct_type>', 17) ->
 12;
yeccgoto('<struct_type>', 24) ->
 133;
yeccgoto('<struct_type>', 117) ->
 133;
yeccgoto('<struct_type>', 198) ->
 133;
yeccgoto('<struct_type>', 207) ->
 12;
yeccgoto('<struct_type>', 208) ->
 12;
yeccgoto('<struct_type>', 251) ->
 12;
yeccgoto('<switch_body>', 54) ->
 56;
yeccgoto('<switch_type_spec>', 28) ->
 39;
yeccgoto('<template_type_spec>', 24) ->
 134;
yeccgoto('<template_type_spec>', 117) ->
 134;
yeccgoto('<template_type_spec>', 152) ->
 134;
yeccgoto('<template_type_spec>', 198) ->
 134;
yeccgoto('<type_dcl>', 0) ->
 13;
yeccgoto('<type_dcl>', 17) ->
 13;
yeccgoto('<type_dcl>', 207) ->
 13;
yeccgoto('<type_dcl>', 208) ->
 13;
yeccgoto('<type_dcl>', 251) ->
 258;
yeccgoto('<type_declarator>', 24) ->
 186;
yeccgoto('<type_spec>', 24) ->
 187;
yeccgoto('<type_spec>', 117) ->
 135;
yeccgoto('<type_spec>', 198) ->
 199;
yeccgoto('<unary_expr>', 64) ->
 84;
yeccgoto('<unary_expr>', 67) ->
 84;
yeccgoto('<unary_expr>', 90) ->
 84;
yeccgoto('<unary_expr>', 92) ->
 84;
yeccgoto('<unary_expr>', 94) ->
 84;
yeccgoto('<unary_expr>', 95) ->
 84;
yeccgoto('<unary_expr>', 97) ->
 84;
yeccgoto('<unary_expr>', 98) ->
 84;
yeccgoto('<unary_expr>', 100) ->
 105;
yeccgoto('<unary_expr>', 101) ->
 104;
yeccgoto('<unary_expr>', 102) ->
 103;
yeccgoto('<unary_expr>', 111) ->
 84;
yeccgoto('<unary_expr>', 145) ->
 84;
yeccgoto('<unary_expr>', 149) ->
 84;
yeccgoto('<unary_expr>', 154) ->
 84;
yeccgoto('<unary_expr>', 165) ->
 84;
yeccgoto('<unary_expr>', 245) ->
 84;
yeccgoto('<unary_operator>', 64) ->
 85;
yeccgoto('<unary_operator>', 67) ->
 85;
yeccgoto('<unary_operator>', 90) ->
 85;
yeccgoto('<unary_operator>', 92) ->
 85;
yeccgoto('<unary_operator>', 94) ->
 85;
yeccgoto('<unary_operator>', 95) ->
 85;
yeccgoto('<unary_operator>', 97) ->
 85;
yeccgoto('<unary_operator>', 98) ->
 85;
yeccgoto('<unary_operator>', 100) ->
 85;
yeccgoto('<unary_operator>', 101) ->
 85;
yeccgoto('<unary_operator>', 102) ->
 85;
yeccgoto('<unary_operator>', 111) ->
 85;
yeccgoto('<unary_operator>', 145) ->
 85;
yeccgoto('<unary_operator>', 149) ->
 85;
yeccgoto('<unary_operator>', 154) ->
 85;
yeccgoto('<unary_operator>', 165) ->
 85;
yeccgoto('<unary_operator>', 245) ->
 85;
yeccgoto('<union_type>', 0) ->
 14;
yeccgoto('<union_type>', 17) ->
 14;
yeccgoto('<union_type>', 24) ->
 136;
yeccgoto('<union_type>', 117) ->
 136;
yeccgoto('<union_type>', 198) ->
 136;
yeccgoto('<union_type>', 207) ->
 14;
yeccgoto('<union_type>', 208) ->
 14;
yeccgoto('<union_type>', 251) ->
 14;
yeccgoto('<unsigned_int>', 18) ->
 40;
yeccgoto('<unsigned_int>', 24) ->
 40;
yeccgoto('<unsigned_int>', 28) ->
 40;
yeccgoto('<unsigned_int>', 117) ->
 40;
yeccgoto('<unsigned_int>', 152) ->
 40;
yeccgoto('<unsigned_int>', 198) ->
 40;
yeccgoto('<unsigned_int>', 261) ->
 40;
yeccgoto('<unsigned_int>', 265) ->
 40;
yeccgoto('<unsigned_int>', 299) ->
 40;
yeccgoto('<unsigned_long_int>', 18) ->
 41;
yeccgoto('<unsigned_long_int>', 24) ->
 41;
yeccgoto('<unsigned_long_int>', 28) ->
 41;
yeccgoto('<unsigned_long_int>', 117) ->
 41;
yeccgoto('<unsigned_long_int>', 152) ->
 41;
yeccgoto('<unsigned_long_int>', 198) ->
 41;
yeccgoto('<unsigned_long_int>', 261) ->
 41;
yeccgoto('<unsigned_long_int>', 265) ->
 41;
yeccgoto('<unsigned_long_int>', 299) ->
 41;
yeccgoto('<unsigned_short_int>', 18) ->
 42;
yeccgoto('<unsigned_short_int>', 24) ->
 42;
yeccgoto('<unsigned_short_int>', 28) ->
 42;
yeccgoto('<unsigned_short_int>', 117) ->
 42;
yeccgoto('<unsigned_short_int>', 152) ->
 42;
yeccgoto('<unsigned_short_int>', 198) ->
 42;
yeccgoto('<unsigned_short_int>', 261) ->
 42;
yeccgoto('<unsigned_short_int>', 265) ->
 42;
yeccgoto('<unsigned_short_int>', 299) ->
 42;
yeccgoto('<xor_expr>', 64) ->
 86;
yeccgoto('<xor_expr>', 67) ->
 86;
yeccgoto('<xor_expr>', 111) ->
 112;
yeccgoto('<xor_expr>', 145) ->
 86;
yeccgoto('<xor_expr>', 149) ->
 86;
yeccgoto('<xor_expr>', 154) ->
 86;
yeccgoto('<xor_expr>', 165) ->
 86;
yeccgoto('<xor_expr>', 245) ->
 86;
yeccgoto('OE_pragma', 0) ->
 15;
yeccgoto('OE_pragma', 17) ->
 15;
yeccgoto('OE_pragma', 58) ->
 60;
yeccgoto('OE_pragma', 62) ->
 60;
yeccgoto('OE_pragma', 116) ->
 60;
yeccgoto('OE_pragma', 117) ->
 60;
yeccgoto('OE_pragma', 169) ->
 60;
yeccgoto('OE_pragma', 171) ->
 60;
yeccgoto('OE_pragma', 172) ->
 60;
yeccgoto('OE_pragma', 198) ->
 60;
yeccgoto('OE_pragma', 201) ->
 60;
yeccgoto('OE_pragma', 203) ->
 60;
yeccgoto('OE_pragma', 207) ->
 15;
yeccgoto('OE_pragma', 208) ->
 15;
yeccgoto('OE_pragma', 221) ->
 60;
yeccgoto('OE_pragma', 227) ->
 60;
yeccgoto('OE_pragma', 230) ->
 60;
yeccgoto('OE_pragma', 232) ->
 60;
yeccgoto('OE_pragma', 235) ->
 60;
yeccgoto('OE_pragma', 251) ->
 259;
yeccgoto('OE_pragma', 297) ->
 60;
yeccgoto('OE_pragma', 304) ->
 60;
yeccgoto('OE_pragma', 307) ->
 60;
yeccgoto('OE_pragma', 309) ->
 60;
yeccgoto('OE_pragma', 311) ->
 60;
yeccgoto('OE_preproc', 0) ->
 16;
yeccgoto('OE_preproc', 17) ->
 16;
yeccgoto('OE_preproc', 207) ->
 16;
yeccgoto('OE_preproc', 208) ->
 16;
yeccgoto('OE_preproc', 251) ->
 260;
yeccgoto('OorM_<case>', 54) ->
 57;
yeccgoto('OorM_<case_label>', 58) ->
 61;
yeccgoto('OorM_<definition>', 0) ->
 17;
yeccgoto('OorM_<definition>', 207) ->
 208;
yeccgoto('OorM_<fixed_array_size>', 161) ->
 164;
yeccgoto('OorM_<member>', 194) ->
 197;
yeccgoto('Opt_<context_expr>', 282) ->
 289;
yeccgoto('Opt_<inheritance_spec>', 211) ->
 214;
yeccgoto('Opt_<op_attribute>', 251) ->
 261;
yeccgoto('Opt_<raises_expr>', 280) ->
 282;
yeccgoto('Opt_readonly', 251) ->
 262;
yeccgoto('Uggly_pragmas', 54) ->
 58;
yeccgoto('Uggly_pragmas', 57) ->
 58;
yeccgoto('Uggly_pragmas', 58) ->
 62;
yeccgoto('Uggly_pragmas', 61) ->
 117;
yeccgoto('Uggly_pragmas', 63) ->
 116;
yeccgoto('Uggly_pragmas', 121) ->
 172;
yeccgoto('Uggly_pragmas', 124) ->
 169;
yeccgoto('Uggly_pragmas', 170) ->
 171;
yeccgoto('Uggly_pragmas', 194) ->
 198;
yeccgoto('Uggly_pragmas', 197) ->
 198;
yeccgoto('Uggly_pragmas', 200) ->
 201;
yeccgoto('Uggly_pragmas', 202) ->
 203;
yeccgoto('Uggly_pragmas', 220) ->
 221;
yeccgoto('Uggly_pragmas', 222) ->
 198;
yeccgoto('Uggly_pragmas', 226) ->
 227;
yeccgoto('Uggly_pragmas', 228) ->
 230;
yeccgoto('Uggly_pragmas', 231) ->
 232;
yeccgoto('Uggly_pragmas', 233) ->
 235;
yeccgoto('Uggly_pragmas', 279) ->
 297;
yeccgoto('Uggly_pragmas', 300) ->
 304;
yeccgoto('Uggly_pragmas', 305) ->
 307;
yeccgoto('Uggly_pragmas', 308) ->
 309;
yeccgoto('Uggly_pragmas', 310) ->
 311;
yeccgoto('ZorM_<declarator>', 188) ->
 190;
yeccgoto('ZorM_<enumerator>', 230) ->
 231;
yeccgoto('ZorM_<export>', 249) ->
 251;
yeccgoto('ZorM_<integer_literal>', 325) ->
 327;
yeccgoto('ZorM_<integer_literal>', 326) ->
 329;
yeccgoto('ZorM_<member>', 220) ->
 222;
yeccgoto('ZorM_<param_dcl>', 300) ->
 305;
yeccgoto('ZorM_<scoped_name>', 215) ->
 216;
yeccgoto('ZorM_<scoped_name>', 285) ->
 286;
yeccgoto('ZorM_<simple_declarator>', 271) ->
 272;
yeccgoto('ZorM_<string_literal>', 292) ->
 293;
yeccgoto(__Symbol, __State) ->
 exit({__Symbol, __State, missing_in_goto_table}).


